// Zustzliche Standard-Header drfen hier eingebunden werden.
#include <iterator>
#include <algorithm>

namespace Darkrel
{
    double computeScore(const PlanetMap& planetMap,
                        const Border& border)
    {
        // zunchst die Lsung auf Gltigkeit prfen
        for(int x = 0; x < MAP_WIDTH; ++x)
        {
            if(border[x] <= 0 || border[x] >= MAP_HEIGHT)
            {
                char msg[256];
                sprintf(msg, "Ungltiger Grenzverlauf! border[%d] = %d", x, border[x]);
                throw std::runtime_error(msg);
            }
        }

        // Die Bewertung ergibt sich aus zwei Teilen:
        // 1.) "Lnge" der Grenze (s.u.)
        // 2.) Qualitt der Aufteilung (s.u.)

        // Zuerst wird die "Lnge" der Grenze ermittelt.
        // Jede y-nderung dy der Grenze von x nach x+1 (und von x=511 nach x=0)
        // fhrt zu einem Punktabzug von dy/2048.
        // Negative Punktzahlen sind mglich!
        // Eine in dieser Hinsicht optimale Grenze ist einfach eine horizontale Linie.
        // Eine solche Grenze bringt genau 1 Punkt.

        int borderLength = 0;
        for(int x = 0; x < MAP_WIDTH - 1; ++x) borderLength += absInt(border[x] - border[x + 1]);
        borderLength += absInt(border[MAP_WIDTH - 1] - border[0]);
        const double borderScore = 1.0 - borderLength / 2048.0;

        // Nun wird die Qualitt der Aufteilung des Planeten bestimmt.
        // Diese besteht aus zwei Faktoren:
        //
        // 1.) Wasser(Wasserbewohner) / Wasser(Gesamt)
        //     = Anteil des nutzbaren Wassers am Gesamtvorkommen auf dem Planeten
        //       (Wasser, das den Landbewohnern zugeteilt wird, gilt als verschwendet)
        //
        // 2.) Land(Landbewohner) / Land(Gesamt)
        //     = Anteil des nutzbaren Lands am Gesamtvorkommen auf dem Planeten
        //       (Land, das den Wasserbewohnern zugeteilt wird, gilt als verschwendet)
        //
        // Diese Faktoren werden miteinander multipliziert
        // und ergeben die zweite Teilpunktzahl.
        // Das theoretische Maximum ist 1, nmlich wenn die eine Hlfte des Planeten
        // zu 100% aus Wasser besteht und die andere zu 100% aus Land.

        // bestimmen, wie viel Wasser/Land nrdlich und sdlich der Grenze vorkommen
        int waterNorth = 0;
        int waterSouth = 0;
        int landNorth = 0;
        int landSouth = 0;
        for(int x = 0; x < MAP_WIDTH; ++x)
        {
            // nrdliche Seite der Grenze
            int landNorthInColumn = 0;
            for(int y = 0; y < border[x]; ++y) landNorthInColumn += planetMap[x][y] ? 1 : 0;
            landNorth += landNorthInColumn;
            waterNorth += border[x] - landNorthInColumn;

            // sdliche Seite der Grenze
            int landSouthInColumn = 0;
            for(int y = border[x]; y < MAP_HEIGHT; ++y) landSouthInColumn += planetMap[x][y] ? 1 : 0;
            landSouth += landSouthInColumn;
            waterSouth += (MAP_HEIGHT - border[x]) - landSouthInColumn;
        }

        // Gesamtvorkommen berechnen
        const int waterTotal = waterNorth + waterSouth;
        const int landTotal = landNorth + landSouth;

        // Wir wissen noch nicht, welche Bewohner auf welcher Seite leben sollen.
        // Darum probieren wir beide Mglichkeiten und whlen dann die bessere.
        const double divisionScore1 = (waterNorth / static_cast<double>(waterTotal)) * (landSouth / static_cast<double>(landTotal));
        const double divisionScore2 = (waterSouth / static_cast<double>(waterTotal)) * (landNorth / static_cast<double>(landTotal));
        const double bestDivisionScore = std::max(divisionScore1, divisionScore2);

        // Die endgltige Punktzahl ergibt sich:
        // - zu 1/3 aus der Punktzahl fr die Lnge der Grenze
        // - zu 2/3 aus der Punktzahl fr die Qualitt der Aufteilung
        return (borderScore + 2.0 * bestDivisionScore) / 3.0;
    }

// Zustzliche Klassen/Funktionen drfen hier definiert werden.
const unsigned int INITIAL_MUTATION_SIZE = 100;

typedef Border Genom;

template<typename T>
inline T clamp(T x, T a, T b){   return x < a ? a : (x > b ? b : x);}

void invert(Genom &genom)
{
	for(int x = 0; x < MAP_WIDTH; ++x)
		genom[x] = (MAP_HEIGHT - 1) - genom[x];
}

///-------------------------------------------------------------------------------------------------
/// <summary>	Perform a random crossover of two Genoms to create a new one. </summary>
///
/// <param name="first">	The first Genom. </param>
/// <param name="second">	The second Genom. </param>
/// <param name="outGenom">	[out] Cross over result of the two other Genoms. </param>
///-------------------------------------------------------------------------------------------------
void crossOver(Genom &first, Genom &second, Genom &outGenom)
{
	int split = (rand() % (MAP_WIDTH - 2)) + 1;
	for(int x = 0; x < MAP_WIDTH; ++x)
		outGenom[x] = x < split ? first[x] : second[x];
}

///-------------------------------------------------------------------------------------------------
/// <summary>	Mutates the genom to a new Genom. </summary>
///
/// <param name="genom">	[out] The genom. </param>
///-------------------------------------------------------------------------------------------------
void mutate(Genom &genom, unsigned int minimal_mutation_size, unsigned int maximal_mutation_size, unsigned int maximum_dy_abs)
{
	int dy = ((rand() % 2) - 1) * (rand() % MAP_WIDTH - 2) + 1;
	dy = clamp<int>(dy, -maximum_dy_abs, maximum_dy_abs);
	int x0 = rand() % MAP_WIDTH;
	int x1 = rand() % MAP_WIDTH;
	while(absInt(x1 - x0) < minimal_mutation_size
		|| absInt(x1 - x0) > maximal_mutation_size)
		x1 = rand() % MAP_WIDTH;
	if(x0 > x1)
	{
		int t = x0;
		x0 = x1;
		x1 = t;
	}

	for(int x = x0; x <= x1; ++x)
	{
		genom[x] =
		genom[x] + dy < 1 ? 1 :
		genom[x] + dy > MAP_HEIGHT - 2 ?  MAP_HEIGHT - 2 : genom[x] + dy;
	}
}

///-------------------------------------------------------------------------------------------------
/// <summary>	Copies the genom. </summary>
///
/// <param name="target">	[in,out] Target for the copy. </param>
/// <param name="source">	[in,out] Source for the copy. </param>
///-------------------------------------------------------------------------------------------------
void copyGenom(Genom &target, Genom &source)
{
	for(int x = 0; x < MAP_WIDTH; ++x)
	    target[x] = source[x];
}

///-------------------------------------------------------------------------------------------------
/// <summary>	Genom sorting. </summary>
///-------------------------------------------------------------------------------------------------
void mergeSort(Genom *&genoms, double *&scores, int l, int r)
{
	if(r - l <= 1) return;

	int m = l + (r - l) / 2;
	mergeSort(genoms, scores, l, m);
	mergeSort(genoms, scores, m+1, r);

	//Merge the two subarrays
	int l1 = l;
	int l2 = m + 1;

	Genom *sorted = new Genom[r - l];
	double *sortedScores = new double[r - l];
	for(int i = 0; i < r - l; ++i)
	{
	    if(l1 == m + 1)
	    {
			copyGenom(sorted[i], genoms[l2]);
			sortedScores[i]  = scores[l2];
			++l2;
	    }
	    else if(l2 == r)
	    {
			copyGenom(sorted[i], genoms[l1]);
			sortedScores[i]  = scores[l1];
			++l1;
	    }
		else if(scores[l1] < scores[l2])
		{
			copyGenom(sorted[i], genoms[l2]);
			sortedScores[i]  = scores[l2];
			++l2;
		}
		else
		{
			copyGenom(sorted[i], genoms[l1]);
			sortedScores[i] = scores[l1];
			++l1;
		}
	}

	for(int i = 0; i < r - l; ++i)
	{
		copyGenom(genoms[i + l], sorted[i]);
		scores[i + l] = sortedScores[i];
	}

	delete[] sorted;
	delete[] sortedScores;
}

///-------------------------------------------------------------------------------------------------
/// <summary>	Makes a constant genom. </summary>
///
/// <param name="genom">	[in,out] The genom. </param>
/// <param name="y">		The y coordinate. </param>
///-------------------------------------------------------------------------------------------------
void makeConstGenom(Genom &genom, int y)
{
	for(int x = 0; x < MAP_WIDTH; ++x)
		genom[x] = y;
}


///-------------------------------------------------------------------------------------------------
/// <summary>	Create the next Generation. </summary>
///
/// <param name="currentPopulation">	[in,out] The current population. </param>
///-------------------------------------------------------------------------------------------------
void nextGeneration(
	Genom *&currentPopulation, const PlanetMap &planetMap, unsigned int &generation,
	unsigned int invertion_intervall, const unsigned int population_size, unsigned int &minimal_mutation_size,
	unsigned int &maximal_mutation_size, unsigned int &maximum_dy_abs
	)
{
	if(minimal_mutation_size == 0) minimal_mutation_size = INITIAL_MUTATION_SIZE;
	if(maximal_mutation_size == 0) maximal_mutation_size = 2*INITIAL_MUTATION_SIZE;

	if(generation % invertion_intervall == 0)
	{
		for(int i = population_size / 2; i < population_size; ++i)
			invert(currentPopulation[i]);
		invertion_intervall = (rand() % 2000) + 100;
	}

	//New Genoms
	Genom *p = new Genom[population_size];
	for(int i = 0; i < population_size; ++i)
	{
		int c = rand() % population_size;

		crossOver(currentPopulation[c], currentPopulation[i], p[i]);
		mutate(p[i], minimal_mutation_size, maximal_mutation_size, maximum_dy_abs);
	}

	int y = clamp(rand() % MAP_HEIGHT, 1, MAP_HEIGHT - 2);
	makeConstGenom(currentPopulation[population_size - 1], y);

	Genom *survival_of_the_fittest = new Genom[2*population_size];
	double *scores = new double[2*population_size];
	for(int i = 0; i < population_size; ++i)
	{
		copyGenom(survival_of_the_fittest[i], currentPopulation[i]);
		copyGenom(survival_of_the_fittest[population_size + i], p[i]);
		scores[i] = computeScore(planetMap, currentPopulation[i]);
		scores[population_size + i] = computeScore(planetMap, p[i]);
	}
	mergeSort(survival_of_the_fittest, scores, 0, 2*population_size);
	for(int i = 0; i < population_size; ++i)
		copyGenom(currentPopulation[i], survival_of_the_fittest[i]);


	minimal_mutation_size = INITIAL_MUTATION_SIZE * (1 - scores[0]);
	maximal_mutation_size = INITIAL_MUTATION_SIZE * scores[0];
	maximum_dy_abs = (1 - scores[0]) * INITIAL_MUTATION_SIZE;

	delete[] p;
	delete[] survival_of_the_fittest;
	delete[] scores;

	++generation;
	makeConstGenom(currentPopulation[population_size - 1], (rand() % (MAP_HEIGHT - 2)) + 1);
}


void dividePlanet(const PlanetMap& planetMap,
                  Border& outBorder,
                  const Clock& clock)
{
	srand(time(NULL));
	int population_size = (rand() % 30) + 61;
	Genom *p = new Genom[population_size];
	for(int i = 0; i < population_size; ++i)
	{
		makeConstGenom(p[i], (i + 1) * (MAP_HEIGHT/(population_size + 2)));
	}

	double averageTime = 0.0F;
	unsigned int generation = 1;
	unsigned int minimal_mutation_size = 1;
	unsigned int maximal_mutation_size = 500;
	unsigned int invertion_intervall = 1000;
	unsigned int maximum_dy_abs = INITIAL_MUTATION_SIZE;
	//20 ms um das maximum rauszuholen. Das ist vermutlich zu viel.
	while(0.999 - clock() > averageTime)
	{
		nextGeneration(p, planetMap, generation, invertion_intervall, population_size, minimal_mutation_size, maximal_mutation_size, maximum_dy_abs);
		averageTime = clock() / generation;
	}

    for(int x = 0; x < MAP_WIDTH; ++x) outBorder[x] = p[0][x];

	delete[] p;

    // mach etwas sinnvolles, so lange noch Zeit brig ist ...
	// Kommentar: Drfte nicht zustande kommen ;)
    while(clock() < 0.99) { }
}

}
