// Zustzliche Standard-Header drfen hier eingebunden werden.
#include <cstring>
#include <climits>
#include <cstdint>
#include <cfloat>

namespace David_Scherfgen
{

/* Mein Algorithmus:
 *
 * 1. Die Planetenkarte wird auf 256x128, 64x32, 32x16, 16x8 und 8x4 herunterskaliert.
 *    Dabei entsteht ein "Graustufenbild", da die Summe der zusammengefassten Pixel gespeichert wird.
 *    Beispiel:
 *
 *              8x4               Verkleinerung   4x2
 *
 *              1 1 0 0 1 1 0 0                   4 1 3 2
 *              1 1 0 1 1 0 1 1   ------------>   2 3 1 0
 *              1 1 1 1 0 0 0 0
 *              0 0 0 1 1 0 0 0
 *
 * 2. Fr die 8x4-Karte kann man in kurzer Zeit mittels Brute Force die optimale Grenze berechnen,
 *    da es nur 3^8 = 6561 Mglichkeiten gibt (3 Mglichkeiten fr jeden x-Wert, da die Grenze nicht bei
 *    Null starten darf).
 *    Bei der 16x8-Karte sind es hingegen schon 7^16 = 33232930569601 Mglichkeiten, was nicht mehr
 *    mit Brute Force machbar ist.
 *
 *	  Meine Annahme ist nun, dass die optimale Grenze einer kleineren Version der Karte auch ungefhr(!)
 *	  der optimalen Grenze fr die Karte in der Originalgre entspricht.
 *
 * 3. Die garantiert optimale 8x4-Grenze wird hochskaliert auf 16x8.
 *    Nun arbeite ich weiter auf der 16x8-Karte.
 *    Es folgen einige Optimierungen der Grenze. Dann wird wieder hochskaliert.
 *    Das geht so weiter, bis ich bei 512x256 angekommen bin.
 *
 *    Die Optimierungsschritte, die zwischendurch erfolgen, sind:
 *    - Versuche, Abschnitte von x1 bis x2 ber die ganze Grenze hinweg anzuheben oder abzusenken.
 *      Wenn sich etwas verbessert, bernimm die Verbesserung.
 *    - Betrachte gleichzeitig jeweils x Nachbarpunkte auf der Grenze und verschiebe sie alle individuell,
 *      jedoch gleichzeitig nach oben oder unten oder setze sie auf den Wert eines ihrer Nachbarn
 *      Wenn sich etwas verbessert, bernimm die Verbesserung.
 *    - Finde lokale optimale Grenzverlufe fr Abschnitte von x1 bis x2,
 *      unter der Annahme, dass der Rest des Grenzverlaufs schon ziemlich gut ist.
 *      Wenn sich etwas verbessert, bernimm die Verbesserung.
 *    - und noch ein paar andere Sachen ...
 *
 *    Die Bewertungsfunktion wurde durch das Verwenden von Lookup-Tabellen/Prfixsummen beschleunigt.
 *    So knnen mehr Tests durchgefhrt werden als wenn man einfach die vorgegebene Bewertungsfunktion
 *    benutzt htte.
 *    Helmut hat dies in seinem Algorithmus zur Perfektion gebracht. Meine Bewertungsfunktion bentigt
 *    hingegen immer noch eine Schleife.
 *
 * 4. Wenn am Ende noch Zeit brig ist, werden die oben genannten Optimierungen immer weiter angewendet,
 *    bis sich nichts mehr verbessert.
 */

typedef double ScoreType;
#define SCORE_CMP_ >

template<int W> struct MinSizeInt;
template<> struct MinSizeInt<1024> { typedef uint8_t MapIntType; };  // ?
template<> struct MinSizeInt<512>  { typedef uint8_t MapIntType; };  // max. 1
template<> struct MinSizeInt<256>  { typedef uint8_t MapIntType; };  // max. 4
template<> struct MinSizeInt<128>  { typedef uint8_t MapIntType; };  // max. 16
template<> struct MinSizeInt<64>   { typedef uint8_t MapIntType; };  // max. 64
template<> struct MinSizeInt<32>   { typedef uint16_t MapIntType; }; // max. 256
template<> struct MinSizeInt<16>   { typedef uint16_t MapIntType; }; // max. 1024
template<> struct MinSizeInt<8>    { typedef uint16_t MapIntType; }; // max. 4096

struct Segment
{
	int firstX;
	int lastX;
	int value;
};

template<int W> struct Helper
{
    typedef int BorderType[W];
    typedef typename MinSizeInt<W>::MapIntType MapIntType;

    static const int WIDTH = W;
    static const int HEIGHT = W / 2;

    Helper(const PlanetMap& planetMap)
        : scaleFactor(1)
        , scaleFactorSq(1)
        , scaleFactorLog(0)
        , scaleFactorSqLog(0)
        , totalWater(0)
        , totalLand(0)
    {
        for(int x = 0; x < MAP_WIDTH; ++x)
            for(int y = 0; y < MAP_HEIGHT; ++y)
                map[x][y] = static_cast<MapIntType>(planetMap[x][y]);

        initData();
    }

    Helper(const Helper<W * 2>& helperX2)
        : scaleFactor(helperX2.scaleFactor * 2)
        , scaleFactorSq(scaleFactor * scaleFactor)
        , scaleFactorLog(helperX2.scaleFactorLog + 1)
        , scaleFactorSqLog(scaleFactorLog * 2)
    {
        for(int x = 0; x < W; ++x)
            for(int y = 0; y < W / 2; ++y)
                map[x][y] =   helperX2.map[x * 2][y * 2]
                            + helperX2.map[x * 2 + 1][y * 2]
                            + helperX2.map[x * 2][y * 2 + 1]
                            + helperX2.map[x * 2 + 1][y * 2 + 1];

        initData();
    }

    void initData()
    {
        for(int x = 0; x <= W; ++x) landIntegral[x][0] = 0;
        for(int y = 0; y <= W / 2; ++y) landIntegral[0][y] = 0;

        for(int x = 1; x <= W; ++x)
        {
            landColumnIntegral[0][x - 1] = 0;

            for(int y = 1; y <= HEIGHT; ++y)
            {
                const int land = map[x - 1][y - 1];

                landIntegral[x][y] =   land
                                     + landIntegral[x][y - 1]
                                     + landIntegral[x - 1][y]
                                     - landIntegral[x - 1][y - 1];

                landColumnIntegral[y][x - 1] = land + landColumnIntegral[y - 1][x - 1];
            }
        }

        rect(W, W / 2, totalWater, totalLand);
        twoOverTotalWaterTimesTotalLand = 2.0 / (static_cast<uint64_t>(totalWater) * totalLand);
    }

    void rect(int width,
              int height,
              int& outWater,
              int& outLand) const
    {
        outLand = landIntegral[width][height];
        outWater = (MAP_WIDTH * MAP_HEIGHT) - outLand;
    }

    void rect(int x,
              int y,
              int width,
              int height,
              int& outWater,
              int& outLand) const
    {
        outLand =   landIntegral[x][y]
                  + landIntegral[x + width][y + height]
                  - landIntegral[x + width][y]
                  - landIntegral[x][y + height];
        outWater = (MAP_WIDTH * MAP_HEIGHT) - outLand;
    }

	int extractSegments(int* p_border,
						Segment* p_outSegments,
						int minSegmentLength) const
	{
		p_outSegments[0].firstX = 0;
		p_outSegments[0].lastX = 0;
		p_outSegments[0].value = p_border[0];
		Segment* p_lastSegment = p_outSegments;

		for(int x = 1; x < W; ++x)
		{
			if(p_border[x & (W - 1)] == p_lastSegment->value) ++(p_lastSegment->lastX);
			else
			{
				if(p_lastSegment->lastX - p_lastSegment->firstX + 1 >= minSegmentLength ||
				   p_lastSegment == p_outSegments)
					++p_lastSegment;

				p_lastSegment->firstX = x;
				p_lastSegment->lastX = x;
				p_lastSegment->value = p_border[x & (W - 1)];
			}
		}

		if(p_lastSegment != p_outSegments)
		{
			if(p_lastSegment->lastX == W + p_outSegments[0].firstX - 1 &&
			   p_lastSegment->value == p_outSegments[0].value)
			{
				p_outSegments[0].firstX = p_lastSegment->firstX;
				p_outSegments[0].lastX += W;
				--p_lastSegment;
			}
			else if(p_outSegments[0].lastX - p_outSegments[0].lastX + 1 < minSegmentLength)
			{
				// erstes Segment ist zu kurz, ersetze es durch das letzte
				p_outSegments[0] = *(p_lastSegment--);
			}
		}

		if(p_lastSegment->lastX - p_lastSegment->firstX + 1 < minSegmentLength)
		{
			// letztes Segment zu kurz
			if(p_lastSegment == p_outSegments) return 0;
			else --p_lastSegment;
		}

		return (p_lastSegment - p_outSegments) + 1;
	}

	mutable struct
	{
		ScoreType* p_score;
		int* p_border;
		bool waterNorth;
		struct
		{
			int borderLength;
			int waterNorth;
			int landNorth;
		} prefix[W];
	} superFast;

	void initSuperFastEval(int* p_border,
						   ScoreType* p_score) const
	{
		superFast.p_score = p_score;
		superFast.p_border = p_border;

        superFast.prefix[0].borderLength = absInt(p_border[1] - p_border[0]);
		for(int x = 1; x < W - 1; ++x) superFast.prefix[x].borderLength = superFast.prefix[x - 1].borderLength + absInt(p_border[x + 1] - p_border[x]);
		superFast.prefix[W - 1].borderLength = superFast.prefix[W - 2].borderLength;

		superFast.prefix[0].landNorth = landColumnIntegral[p_border[0]][0];
		superFast.prefix[0].waterNorth = (p_border[0] << scaleFactorSqLog) - superFast.prefix[0].landNorth;
		for(int x = 1; x < W; ++x)
		{
            const int landNorthInColumn = landColumnIntegral[p_border[x]][x];
            superFast.prefix[x].landNorth = superFast.prefix[x - 1].landNorth + landNorthInColumn;
            const int waterNorthInColumn = (p_border[x] << scaleFactorSqLog) - landNorthInColumn;
            superFast.prefix[x].waterNorth = superFast.prefix[x - 1].waterNorth + waterNorthInColumn;
		}
	}

	void reinitSuperFastEval(int firstX) const
	{
	    // Die "Super Fast"-Datenstruktur muss erst ab der Stelle der ersten nderung neu initialisiert werden.

	    const int* p_border = superFast.p_border;

	    if(firstX <= 1)
	    {
	        superFast.prefix[0].borderLength = absInt(p_border[1] - p_border[0]);
            if(firstX == 0)
            {
                superFast.prefix[0].landNorth = landColumnIntegral[p_border[0]][0];
                superFast.prefix[0].waterNorth = (p_border[0] << scaleFactorSqLog) - superFast.prefix[0].landNorth;
            }
	    }

		for(int x = std::max(1, firstX - 1); x < W - 1; ++x) superFast.prefix[x].borderLength = superFast.prefix[x - 1].borderLength + absInt(p_border[x + 1] - p_border[x]);
		superFast.prefix[W - 1].borderLength = superFast.prefix[W - 2].borderLength;

		for(int x = std::max(1, firstX); x < W; ++x)
		{
            const int landNorthInColumn = landColumnIntegral[p_border[x]][x];
            superFast.prefix[x].landNorth = superFast.prefix[x - 1].landNorth + landNorthInColumn;
            const int waterNorthInColumn = (p_border[x] << scaleFactorSqLog) - landNorthInColumn;
            superFast.prefix[x].waterNorth = superFast.prefix[x - 1].waterNorth + waterNorthInColumn;
		}
	}

#define COMPUTE_SCORE_ \
    const ScoreType borderScore = 1.0 - (borderLength << scaleFactorLog) * (1.0 / 2048.0); \
    const ScoreType s1 = static_cast<ScoreType>(waterNorth) * twoOverTotalWaterTimesTotalLand * (totalLand - landNorth); \
    const ScoreType s2 = static_cast<ScoreType>(landNorth) * twoOverTotalWaterTimesTotalLand * (totalWater - waterNorth); \
    const ScoreType score = borderScore + std::max(s1, s2);

	bool superFastSetSingle(int x,
							int y) const
	{
		int borderLength;
		int waterNorth;
		int landNorth;

		if(x == 0)
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[0].borderLength;
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[0].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[0].landNorth;
		}
		else if(x == 1)
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[1].borderLength + absInt(superFast.p_border[0] - superFast.p_border[W - 1]);
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[1].waterNorth + superFast.prefix[0].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[1].landNorth + superFast.prefix[0].landNorth;
		}
		else if(x == W - 1)
		{
			borderLength = superFast.prefix[W - 3].borderLength;
			waterNorth = superFast.prefix[W - 2].waterNorth;
			landNorth = superFast.prefix[W - 2].landNorth;
		}
		else
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x].borderLength + superFast.prefix[x - 2].borderLength + absInt(superFast.p_border[0] - superFast.p_border[W - 1]);
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x].waterNorth + superFast.prefix[x - 1].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x].landNorth + superFast.prefix[x - 1].landNorth;
		}

		borderLength +=   absInt(y - superFast.p_border[(x - 1 + W) & (W - 1)])
                        + absInt(y - superFast.p_border[(x + 1) & (W - 1)]);

		const int landNorthInColumn = landColumnIntegral[y][x];
		landNorth += landNorthInColumn;
		waterNorth += (y << scaleFactorSqLog) - landNorthInColumn;

		COMPUTE_SCORE_

#if defined(_DEBUG)
		BorderType temp;
		memcpy(temp, superFast.p_border, sizeof(BorderType));
		temp[x] = y;
		const ScoreType correctScore = quickEval(temp);
		assert(score == correctScore);
#endif

		if(score SCORE_CMP_ *superFast.p_score)
		{
			*superFast.p_score = score;
			superFast.p_border[x] = y;
			superFast.waterNorth = s1 > s2;
			reinitSuperFastEval(x);
			return true;
		}

		return false;
	}

	bool superFastSetSingleRange(int x,
                                 int y1,
                                 int y2,
                                 int ignoreY = 0) const
	{
#if defined(_DEBUG)
		assert(y1 >= 1 && y1 <= W / 2 - 1);
		assert(y2 >= 1 && y2 <= W / 2 - 1);
		assert(y1 <= y2);
#endif

		int borderLength;
		int waterNorth;
		int landNorth;

		if(x == 0)
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[0].borderLength;
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[0].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[0].landNorth;
		}
		else if(x == 1)
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[1].borderLength + absInt(superFast.p_border[0] - superFast.p_border[W - 1]);
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[1].waterNorth + superFast.prefix[0].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[1].landNorth + superFast.prefix[0].landNorth;
		}
		else if(x == W - 1)
		{
			borderLength = superFast.prefix[W - 3].borderLength;
			waterNorth = superFast.prefix[W - 2].waterNorth;
			landNorth = superFast.prefix[W - 2].landNorth;
		}
		else
		{
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x].borderLength + superFast.prefix[x - 2].borderLength + absInt(superFast.p_border[0] - superFast.p_border[W - 1]);
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x].waterNorth + superFast.prefix[x - 1].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x].landNorth + superFast.prefix[x - 1].landNorth;
		}

		const int startBorderLength = borderLength;
		const int startWaterNorth = waterNorth;
		const int startLandNorth = landNorth;

        int bestY = 0;
        for(int y = y1; y <= y2; ++y)
        {
            if(y == ignoreY) continue;

            borderLength = startBorderLength + absInt(y - superFast.p_border[(x - 1 + W) & (W - 1)])
                                             + absInt(y - superFast.p_border[(x + 1) & (W - 1)]);

            const int landNorthInColumn = landColumnIntegral[y][x];
            landNorth = startLandNorth + landNorthInColumn;
            waterNorth = startWaterNorth + (y << scaleFactorSqLog) - landNorthInColumn;

            COMPUTE_SCORE_

#if defined(_DEBUG)
            BorderType temp;
            memcpy(temp, superFast.p_border, sizeof(BorderType));
            temp[x] = y;
            const ScoreType correctScore = quickEval(temp);
            assert(score == correctScore);
#endif

            if(score SCORE_CMP_ *superFast.p_score)
			{
				*superFast.p_score = score;
				superFast.waterNorth = s1 > s2;
				bestY = y;
			}
        }

        if(bestY > 0)
        {
            superFast.p_border[x] = bestY;
            reinitSuperFastEval(x);
            return true;
        }

        return false;
	}

	bool superFastChangeBlock(int x1,
							  int x2,
							  int y,
							  bool relative = false) const
	{
#if defined(_DEBUG)
		assert(relative || (y >= 1 && y <= W / 2 - 1));
		assert(x2 - x1 + 1 < W);
#endif

	    if(x1 == x2)
	    {
			x1 &= (W - 1);
	        if(relative) y += superFast.p_border[x1];
	        if(y < 1) y = 1;
	        else if(y > W / 2 - 1) y = W / 2 - 1;
	        return superFastSetSingle(x1, y);
	    }

		int borderLength;
		int waterNorth;
		int landNorth;

		if(x1 > 0)
		{
			if(x2 < W) // "Netter" Block, kein Modulo!
			{
				// links vom Block
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth;

				if(x2 < W - 1)
				{
					// rechts vom Block
					borderLength += superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
					waterNorth += superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
					landNorth += superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
				}
			}
			else // Block geht rechts ber die Grenze hinaus.
			{
				const int left = x2 & (W - 1);
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength - superFast.prefix[left].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth - superFast.prefix[left].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth - superFast.prefix[left].landNorth;
			}
		}
		else // Block fngt ganz links an, kein Grenzbertritt mglich!
		{
			// rechts vom Block
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
		}

		if(relative)
		{
			int finalBorderLeft;
			int finalBorderRight;
			if(x1 == 0 || x2 >= W)
			{
				finalBorderLeft = superFast.p_border[0] + y;
				if(finalBorderLeft < 1) finalBorderLeft = 1;
				else if(finalBorderLeft > W / 2 - 1) finalBorderLeft = W / 2 - 1;
			}
			else finalBorderLeft = superFast.p_border[0];
			if(x2 >= W)
			{
				finalBorderRight = superFast.p_border[W - 1] + y;
				if(finalBorderRight < 1) finalBorderRight = 1;
				else if(finalBorderRight > W / 2 - 1) finalBorderRight = W / 2 - 1;
			}
			else finalBorderRight = superFast.p_border[W - 1];

			int left = superFast.p_border[(x1 - 1 + W) & (W - 1)];
			int newBorder = 0;

			for(int i = x1; i <= x2; ++i)
			{
				const int x = i & (W - 1);
				newBorder = superFast.p_border[x] + y;
				if(newBorder < 1) newBorder = 1;
				else if(newBorder > W / 2 - 1) newBorder = W / 2 - 1;

				const int landNorthInColumn = landColumnIntegral[newBorder][x];

				borderLength += absInt(newBorder - left);
				left = newBorder;

				landNorth += landNorthInColumn;
				waterNorth += (newBorder << scaleFactorSqLog) - landNorthInColumn;
			}

			borderLength += absInt(superFast.p_border[(x2 + 1) & (W - 1)] - newBorder);
			if(x1 != 0 && x2 < W - 1) borderLength += absInt(finalBorderLeft - finalBorderRight);
		}
		else
		{
			for(int i = x1; i <= x2; ++i)
			{
				const int landNorthInColumn = landColumnIntegral[y][i & (W - 1)];
				landNorth += landNorthInColumn;

				const int waterNorthInColumn = (y << scaleFactorSqLog) - landNorthInColumn;
				waterNorth += waterNorthInColumn;
			}

			borderLength += absInt(superFast.p_border[(x1 - 1 + W) & (W - 1)] - y);
			borderLength += absInt(superFast.p_border[(x2 + 1) & (W - 1)] - y);
			const int finalBorderLeft = (x1 == 0 || x2 >= W) ? y : superFast.p_border[0];
			const int finalBorderRight = (x2 >= W) ? y : superFast.p_border[W - 1];
			if(x1 != 0 && x2 < W - 1) borderLength += absInt(finalBorderLeft - finalBorderRight);
		}

        COMPUTE_SCORE_

#if defined(_DEBUG)
		BorderType temp;
		memcpy(temp, superFast.p_border, sizeof(BorderType));
		for(int i = x1; i <= x2; ++i)
		{
			const int x = i & (W - 1);
			if(relative)
			{
				temp[x] += y;
				if(temp[x] > W / 2 - 1) temp[x] = W / 2 - 1;
				else if(temp[x] < 1) temp[x] = 1;
			}
			else temp[x] = y;
		}

		const ScoreType correctScore = quickEval(temp);
		assert(score == correctScore);
#endif

		if(score SCORE_CMP_ *superFast.p_score)
		{
			*superFast.p_score = score;
			superFast.waterNorth = s1 > s2;

			if(relative)
			{
				for(int i = x1; i <= x2; ++i)
				{
					const int x = i & (W - 1);
					int newBorder = superFast.p_border[x] + y;
					if(newBorder < 1) newBorder = 1;
					else if(newBorder > W / 2 - 1) newBorder = W / 2 - 1;
					superFast.p_border[x] = newBorder;
				}
			}
			else for(int i = x1; i <= x2; ++i) superFast.p_border[i & (W - 1)] = y;

			reinitSuperFastEval(x2 >= W ? 0 : x1);

			return true;
		}

		return false;
	}

	bool superFastChangeBlockRelativeNoClip(int x1,
                                            int x2,
                                            int y) const
	{
#if defined(_DEBUG)
		assert(x2 - x1 + 1 < W);
#endif

	    if(x1 == x2)
	    {
			x1 &= (W - 1);
	        return superFastSetSingle(x1, superFast.p_border[x1] + y);
	    }

		int borderLength = superFast.prefix[W - 1].borderLength;
		borderLength += absInt(superFast.p_border[W - 1] - superFast.p_border[0]);

		borderLength -= absInt(superFast.p_border[(x1 - 1 + W) & (W - 1)] - superFast.p_border[x1]);
		borderLength -= absInt(superFast.p_border[x2 & (W - 1)] - superFast.p_border[(x2 + 1) & (W - 1)]);
		borderLength += absInt(superFast.p_border[(x1 - 1 + W) & (W - 1)] - (superFast.p_border[x1] + y));
		borderLength += absInt((superFast.p_border[x2 & (W - 1)] + y) - superFast.p_border[(x2 + 1) & (W - 1)]);

		int waterNorth;
		int landNorth;

		if(x1 > 0)
		{
			if(x2 < W) // "Netter" Block, kein Modulo!
			{
				// links vom Block
				waterNorth = superFast.prefix[x1 - 1].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth;

				if(x2 < W - 1)
				{
					waterNorth += superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
					landNorth += superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
				}
			}
			else // Block geht rechts ber die Grenze hinaus.
			{
				const int left = x2 & (W - 1);
				waterNorth = superFast.prefix[x1 - 1].waterNorth - superFast.prefix[left].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth - superFast.prefix[left].landNorth;
			}
		}
		else // Block fngt ganz links an, kein Grenzbertritt mglich!
		{
			// rechts vom Block
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
		}

        int newBorder = 0;
        for(int i = x1; i <= x2; ++i)
        {
            const int x = i & (W - 1);
            newBorder = superFast.p_border[x] + y;
            const int landNorthInColumn = landColumnIntegral[newBorder][x];
            landNorth += landNorthInColumn;
            waterNorth += (newBorder << scaleFactorSqLog) - landNorthInColumn;
        }

        COMPUTE_SCORE_

#if defined(_DEBUG)
		BorderType temp;
		memcpy(temp, superFast.p_border, sizeof(BorderType));
		for(int i = x1; i <= x2; ++i)
		{
			const int x = i & (W - 1);
			temp[x] += y;
		}

		const ScoreType correctScore = quickEval(temp);
		assert(score == correctScore);
#endif

		if(score SCORE_CMP_ *superFast.p_score)
		{
			*superFast.p_score = score;
			superFast.waterNorth = s1 > s2;
			for(int i = x1; i <= x2; ++i) superFast.p_border[i & (W - 1)] += y;
			reinitSuperFastEval(x2 >= W ? 0 : x1);
			return true;
		}

		return false;
	}

	bool superFastReplaceBlock(int x1,
							   int x2,
							   const int* p_replacement) const
	{
	    if(x1 == x2) return superFastSetSingle(x1 & (W - 1), p_replacement[x1 & (W - 1)]);

#if defined(_DEBUG)
		assert(x2 - x1 + 1 < W);
#endif

		int borderLength;
		int waterNorth;
		int landNorth;

		if(x1 > 0)
		{
			if(x2 < W) // "Netter" Block, kein Modulo!
			{
				// links vom Block
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth;

				if(x2 < W - 1)
				{
					// rechts vom Block
					borderLength += superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
					waterNorth += superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
					landNorth += superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
				}
			}
			else // Block geht rechts ber die Grenze hinaus.
			{
				const int left = x2 & (W - 1);
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength - superFast.prefix[left].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth - superFast.prefix[left].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth - superFast.prefix[left].landNorth;
			}
		}
		else // Block fngt ganz links an, kein Grenzbertritt mglich!
		{
			// rechts vom Block
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
		}

		int left = superFast.p_border[(x1 - 1 + W) & (W - 1)];

		for(int i = x1; i <= x2; ++i)
		{
			const int x = i & (W - 1);
			const int y = p_replacement[x];

			borderLength += absInt(y - left);
			left = y;

			const int landNorthInColumn = landColumnIntegral[y][x];
			landNorth += landNorthInColumn;

			const int waterNorthInColumn = (y << scaleFactorSqLog) - landNorthInColumn;
			waterNorth += waterNorthInColumn;
		}

		borderLength += absInt(superFast.p_border[(x2 + 1) & (W - 1)] - p_replacement[x2 & (W - 1)]);
		const int finalBorderLeft = (x1 == 0 || x2 >= W) ? p_replacement[0] : superFast.p_border[0];
		const int finalBorderRight = (x2 >= W) ? p_replacement[W - 1] : superFast.p_border[W - 1];
		if(x1 != 0 && x2 < W - 1) borderLength += absInt(finalBorderLeft - finalBorderRight);

        COMPUTE_SCORE_

#if defined(_DEBUG)
		BorderType temp;
		memcpy(temp, superFast.p_border, sizeof(BorderType));
		for(int i = x1; i <= x2; ++i)
		{
			const int x = i & (W - 1);
			temp[x] = p_replacement[x];
		}

		const ScoreType correctScore = quickEval(temp);
		assert(score == correctScore);
#endif

		if(score SCORE_CMP_ *superFast.p_score)
		{
			*superFast.p_score = score;
			superFast.waterNorth = s1 > s2;

			for(int i = x1; i <= x2; ++i)
			{
				const int x = i & (W - 1);
				superFast.p_border[x] = p_replacement[x];
			}

			reinitSuperFastEval(x2 >= W ? 0 : x1);

			return true;
		}

		return false;
	}

	bool superFastSetBlockRange(int x1,
                                int x2,
                                int y1,
                                int y2,
                                int ignoreY = 0) const
	{
#if defined(_DEBUG)
		assert(y1 >= 1 && y1 <= W / 2 - 1);
		assert(y2 >= 1 && y2 <= W / 2 - 1);
		assert(y1 <= y2);
		assert(x2 - x1 + 1 < W);
#endif

	    if(x1 == x2) return superFastSetSingleRange(x1 & (W - 1), y1, y2, ignoreY);

		int borderLength;
		int waterNorth;
		int landNorth;

		if(x1 > 0)
		{
			if(x2 < W) // "Netter" Block, kein Modulo!
			{
				// links vom Block
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth;

				if(x2 < W - 1)
				{
					// rechts vom Block
					borderLength += superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
					waterNorth += superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
					landNorth += superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
				}
			}
			else // Block geht rechts ber die Grenze hinaus.
			{
				const int left = x2 & (W - 1);
				if(x1 > 1) borderLength = superFast.prefix[x1 - 2].borderLength - superFast.prefix[left].borderLength;
				else borderLength = 0;
				waterNorth = superFast.prefix[x1 - 1].waterNorth - superFast.prefix[left].waterNorth;
				landNorth = superFast.prefix[x1 - 1].landNorth - superFast.prefix[left].landNorth;
			}
		}
		else // Block fngt ganz links an, kein Grenzbertritt mglich!
		{
			// rechts vom Block
			borderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
			waterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
			landNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
		}

        const int startBorderLength = borderLength;
        const int startWaterNorth = waterNorth;
        const int startLandNorth = landNorth;

        int bestY = 0;
        for(int y = y1; y <= y2; ++y)
        {
            if(y == ignoreY) continue;

            borderLength = startBorderLength;
            waterNorth = startWaterNorth;
            landNorth = startLandNorth;

            for(int i = x1; i <= x2; ++i)
            {
                const int landNorthInColumn = landColumnIntegral[y][i & (W - 1)];
                landNorth += landNorthInColumn;

                const int waterNorthInColumn = (y << scaleFactorSqLog) - landNorthInColumn;
                waterNorth += waterNorthInColumn;
            }

            borderLength += absInt(superFast.p_border[(x1 - 1 + W) & (W - 1)] - y);
            borderLength += absInt(superFast.p_border[(x2 + 1) & (W - 1)] - y);
            const int finalBorderLeft = (x1 == 0 || x2 >= W) ? y : superFast.p_border[0];
            const int finalBorderRight = (x2 >= W) ? y : superFast.p_border[W - 1];
            if(x1 != 0 && x2 < W - 1) borderLength += absInt(finalBorderLeft - finalBorderRight);

            COMPUTE_SCORE_

#if defined(_DEBUG)
            BorderType temp;
            memcpy(temp, superFast.p_border, sizeof(BorderType));
            for(int i = x1; i <= x2; ++i) temp[i & (W - 1)] = y;
            const ScoreType correctScore = quickEval(temp);
            assert(score == correctScore);
#endif

            if(score SCORE_CMP_ *superFast.p_score)
			{
				*superFast.p_score = score;
				superFast.waterNorth = s1 > s2;
				bestY = y;
			}
        }

        if(bestY > 0)
        {
            for(int i = x1; i <= x2; ++i) superFast.p_border[i & (W - 1)] = bestY;
			reinitSuperFastEval(x2 >= W ? 0 : x1);
			return true;
		}

		return false;
	}

    ScoreType quickEval(const BorderType& border,
                        bool* p_outWaterNorth = 0) const
    {
        int borderLength = absInt(border[W - 1] - border[0]);

        int landNorth = landColumnIntegral[border[W - 1]][W - 1];
		int waterNorth = (border[W - 1] << scaleFactorSqLog) - landNorth;

		for(int x = 0; x < W - 1; ++x)
		{
			borderLength += absInt(border[x] - border[x + 1]);

            const int landNorthInColumn = landColumnIntegral[border[x]][x];
            landNorth += landNorthInColumn;
            waterNorth += (border[x] << scaleFactorSqLog) - landNorthInColumn;
		}

		const ScoreType borderScore = 1.0 - (borderLength << scaleFactorLog) / 2048.0;

        const int waterSouth = totalWater - waterNorth;
        const int landSouth = totalLand - landNorth;

        const uint64_t s1 = static_cast<uint64_t>(waterNorth) * landSouth;
        const uint64_t s2 = static_cast<uint64_t>(waterSouth) * landNorth;
        const ScoreType bestDivisionScore = std::max(s1, s2) * twoOverTotalWaterTimesTotalLand;
        if(p_outWaterNorth) *p_outWaterNorth = s1 > s2;

        return borderScore + bestDivisionScore;
    }

    bool optimize(int maxNumIterations = INT_MAX,
                  const Clock* p_clock = 0,
				  double timeLimit = 0.0) const
    {
		const ScoreType startScore = *superFast.p_score;
		const int threshold = scaleFactorSq / 2;
		Segment segments[W];
		int newBorder[W];

        while(maxNumIterations-- && (!p_clock || (*p_clock)() < timeLimit))
        {
			const ScoreType oldBestScore = *superFast.p_score;

			// Schritt 1:
			// Versuche jedes border[x] durch Werte im Intervall der Nachbarschaft zu ersetzen.
			// -> verschiebt Anstiege, verfeinert Grenzlinie

            {
                int b[3] = { superFast.p_border[W - 1], superFast.p_border[0], superFast.p_border[1] };
                for(int x = 0; x < W; b[0] = b[1], b[1] = b[2], b[2] = superFast.p_border[(++x + 1) & (W - 1)])
                {
                    const int y1 = std::min(std::min(b[0], b[1]), b[2]);
                    const int y2 = std::max(std::max(b[0], b[1]), b[2]);
                    superFastSetSingleRange(x, y1, y2, b[1]);
                }
            }

            {
                int b[3] = { superFast.p_border[W - 2], superFast.p_border[W - 1], superFast.p_border[0] };
                for(int x = W - 1; x >= 0; b[2] = b[1], b[1] = b[0], b[0] = superFast.p_border[(--x - 1 + W) & (W - 1)])
                {
                    const int y1 = std::min(std::min(b[0], b[1]), b[2]);
                    const int y2 = std::max(std::max(b[0], b[1]), b[2]);
                    superFastSetSingleRange(x, y1, y2, b[1]);
                }
            }

			// Schritt 2:
			// Segmente finden und einzeln testweise nach oben/unten verschieben
			// (Segment = Abschnitt gleicher Grenzhhe)

			int numSegments = extractSegments(superFast.p_border, segments, 1);
			if(numSegments > 1)
			{
				for(const Segment* p_seg = segments;
					p_seg < segments + numSegments;
					++p_seg)
				{
					const int y1 = std::max(1, p_seg->value - 2);
					const int y2 = std::min(W / 2 - 1, p_seg->value + 2);
					superFastSetBlockRange(p_seg->firstX, p_seg->lastX, y1, y2, p_seg->value);
				}
			}

			// Schritt 3:
			// Segmente abflachen

			numSegments = extractSegments(superFast.p_border, segments, 1);
			if(numSegments > 1)
			{
				for(const Segment* p_seg = segments;
					p_seg < segments + numSegments;
					++p_seg)
				{
					const int yLeft = superFast.p_border[(p_seg->firstX - 1 + W) & (W - 1)];
					if(yLeft > p_seg->value)
					{
						int y = yLeft;
						for(int i = p_seg->firstX; i <= p_seg->lastX; ++i)
						{
							const int x = i & (W - 1);
							if(superFast.waterNorth)
							{
								while(y > p_seg->value && y > 1 && map[x][y - 1] > threshold) --y;
								newBorder[x] = y;
								superFastReplaceBlock(p_seg->firstX, i, newBorder);
							}
							else
							{
								while(y > p_seg->value && y > 1 && map[x][y - 1] <= threshold) --y;
								newBorder[x] = y;
								superFastReplaceBlock(p_seg->firstX, i, newBorder);
							}
						}
					}
					else
					{
						int y = yLeft;
						for(int i = p_seg->firstX; i <= p_seg->lastX; ++i)
						{
							const int x = i & (W - 1);
							if(superFast.waterNorth)
							{
								while(y < p_seg->value && y < (W / 2) - 1 && map[x][y] <= threshold) ++y;
								newBorder[x] = y;
								superFastReplaceBlock(p_seg->firstX, i, newBorder);
							}
							else
							{
								while(y < p_seg->value && y < (W / 2) - 1 && map[x][y] > threshold) ++y;
								newBorder[x] = y;
								superFastReplaceBlock(p_seg->firstX, i, newBorder);
							}
						}
					}
				}
			}

			// abbrechen, wenn sich nichts mehr gendert hat
			if(*superFast.p_score == oldBestScore) break;
        }

		return *superFast.p_score > startScore;
    }

	bool optimizeLocalBinary2(int maxNumCombinations,
							  const Clock* p_clock = 0,
							  double timeLimit = 0.0) const
	{
		const ScoreType startScore = *superFast.p_score;

		int candidatesY[W][6];
		const int* p_border = superFast.p_border;
		for(int x = 0; x < W; ++x)
		{
			const int y = p_border[x];
			int n = 0;
			candidatesY[x][++n] = y;
			if(y > 1) candidatesY[x][++n] = y - 1;
			if(y < (W / 2) - 1) candidatesY[x][++n] = (W / 2) - 1;
			const int rightY = p_border[(x + 1) & (W - 1)];
			if(rightY < y - 1 || rightY > y + 1) candidatesY[x][++n] = rightY;
			const int leftY = p_border[(x - 1 + W) & (W - 1)];
			if(leftY < y - 1 || leftY > y + 1) candidatesY[x][++n] = leftY;
			candidatesY[x][0] = n;
		}

		BorderType newBorder;
		int firstX = 0;
		while(firstX < W)
		{
			if(p_clock && (*p_clock)() > timeLimit) break;

			// Wie viele x-Werte knnen wir gleichzeitig optimieren?
			int numCombinations = 1;
			int lastX = firstX;
			while(lastX < W - 1 && numCombinations * candidatesY[lastX][0] <= maxNumCombinations) numCombinations *= candidatesY[lastX++][0];

			// Let's do it!
			for(int c = 0; c < numCombinations; ++c)
			{
				int n = c;
				for(int x = firstX; x <= lastX; ++x)
				{
					const int numCandidates = candidatesY[x][0];
					switch(numCandidates)
					{
					case 5:
						newBorder[x] = candidatesY[x][1 + (n % 5)];
						n /= 5;
						break;
					case 4:
						newBorder[x] = candidatesY[x][1 + (n & 3)];
						n >>= 2;
						break;
					case 3:
						newBorder[x] = candidatesY[x][1 + (n % 3)];
						n /= 3;
						break;
					case 2:
						newBorder[x] = candidatesY[x][1 + (n & 1)];
						n >>= 1;
						break;
					case 1:
						newBorder[x] = candidatesY[x][1];
						break;
					}
				}

				superFastReplaceBlock(firstX, lastX, newBorder);
			}

			firstX = ((firstX + lastX) >> 1) + 1;
		}

		return *superFast.p_score > startScore;
	}

	bool optimizeLocalBlock(int x1,
							int x2,
							int spreadY) const
	{
		int preBorderLength;
		int preWaterNorth;
		int preLandNorth;

		if(x1 > 0)
		{
			if(x2 < W)
			{
				if(x1 > 1) preBorderLength = superFast.prefix[x1 - 2].borderLength;
				else preBorderLength = 0;
				preWaterNorth = superFast.prefix[x1 - 1].waterNorth;
				preLandNorth = superFast.prefix[x1 - 1].landNorth;

				if(x2 < W - 1)
				{
					preBorderLength += superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
					preWaterNorth += superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
					preLandNorth += superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
				}
			}
			else
			{
				const int left = x2 & (W - 1);
				if(x1 > 1) preBorderLength = superFast.prefix[x1 - 2].borderLength - superFast.prefix[left].borderLength;
				else preBorderLength = 0;
				preWaterNorth = superFast.prefix[x1 - 1].waterNorth - superFast.prefix[left].waterNorth;
				preLandNorth = superFast.prefix[x1 - 1].landNorth - superFast.prefix[left].landNorth;
			}
		}
		else
		{
			preBorderLength = superFast.prefix[W - 1].borderLength - superFast.prefix[x2].borderLength;
			preWaterNorth = superFast.prefix[W - 1].waterNorth - superFast.prefix[x2].waterNorth;
			preLandNorth = superFast.prefix[W - 1].landNorth - superFast.prefix[x2].landNorth;
		}

		const int oldPreBorderLength = preBorderLength;
		const int oldPreWaterNorth = preWaterNorth;
		const int oldPreLandNorth = preLandNorth;

		int replacedBorder[W];
		int left = superFast.p_border[(x1 - 1 + W) & (W - 1)];
		const int rightAfterBlock = superFast.p_border[(x2 + 1) & (W - 1)];
        bool changed = false;

		for(int i = x1; i <= x2; ++i)
		{
			const int x = i & (W - 1);
			const int y1 = std::max(1, superFast.p_border[x] - spreadY);
			const int y2 = std::min(W / 2 - 1, superFast.p_border[x] + spreadY);

			ScoreType bestScore = -DBL_MAX;
			int bestY = 0;

			for(int y = y1; y <= y2; ++y)
			{
				const int borderLength = preBorderLength + absInt(y - left) + absInt(y - rightAfterBlock);
				const int landNorthInColumn = landColumnIntegral[y][x];
				const int landNorth = preLandNorth + landNorthInColumn;
				const int waterNorth = preWaterNorth + (y << scaleFactorSqLog) - landNorthInColumn;

				COMPUTE_SCORE_

				if(score > bestScore)
				{
					bestY = y;
					bestScore = score;
				}
			}

			preBorderLength += absInt(bestY - left);
			const int landNorthInColumn = landColumnIntegral[bestY][x];
			preLandNorth += landNorthInColumn;
			preWaterNorth += (bestY << scaleFactorSqLog) - landNorthInColumn;

			replacedBorder[x] = bestY;
			left = bestY;
			changed |= (bestY != superFast.p_border[x]);
		}

		bool improved = false;
		if(changed) improved |= superFastReplaceBlock(x1, x2, replacedBorder);

		preBorderLength = oldPreBorderLength;
		preWaterNorth = oldPreWaterNorth;
		preLandNorth = oldPreLandNorth;

		int right = superFast.p_border[(x2 + 1) & (W - 1)];
		const int leftBeforeBlock = superFast.p_border[(x1 - 1 + W) & (W - 1)];
        changed = false;

		for(int i = x2; i >= x1; --i)
		{
			const int x = i & (W - 1);
			const int y1 = std::max(1, superFast.p_border[x] - spreadY);
			const int y2 = std::min(W / 2 - 1, superFast.p_border[x] + spreadY);

			ScoreType bestScore = -DBL_MAX;
			int bestY = 0;

			for(int y = y1; y <= y2; ++y)
			{
				const int borderLength = preBorderLength + absInt(y - right) + absInt(y - leftBeforeBlock);
				const int landNorthInColumn = landColumnIntegral[y][x];
				const int landNorth = preLandNorth + landNorthInColumn;
				const int waterNorth = preWaterNorth + (y << scaleFactorSqLog) - landNorthInColumn;

				COMPUTE_SCORE_

				if(score > bestScore)
				{
					bestY = y;
					bestScore = score;
				}
			}

			preBorderLength += absInt(bestY - right);
			const int landNorthInColumn = landColumnIntegral[bestY][x];
			preLandNorth += landNorthInColumn;
			preWaterNorth += (bestY << scaleFactorSqLog) - landNorthInColumn;

			replacedBorder[x] = bestY;
			right = bestY;
			changed |= (bestY != superFast.p_border[x]);
		}

		if(changed) improved |= superFastReplaceBlock(x1, x2, replacedBorder);
		return improved;
	}

	bool optimizeLocalBlocks2(int minBlockWidth,
                              int maxBlockWidth,
                              int blockWidthStep,
                              int spreadY,
                              const Clock* p_clock = 0,
							  double timeLimit = 0.0) const
	{
		const ScoreType startScore = *superFast.p_score;

        for(int w = minBlockWidth; w <= maxBlockWidth; w += blockWidthStep)
        {
            for(int x = 0; x < W; ++x)
            {
                if(p_clock && !(x & 3) && (*p_clock)() > timeLimit) goto abortAlgorithm;
                optimizeLocalBlock(x, x + w - 1, spreadY);
            }
        }

    abortAlgorithm:
		return *superFast.p_score > startScore;
	}

	bool optimizeLocalBlocks3(int minBlockWidth,
                              int maxBlockWidth,
                              int blockWidthStep,
                              int spreadY,
                              const Clock& clock,
							  double timeLimitSuccess,
							  double timeLimitOtherwise) const
	{
		const ScoreType startScore = *superFast.p_score;
		bool improved = false;

        for(int w = minBlockWidth; w <= maxBlockWidth; w += blockWidthStep)
        {
            for(int x = 0; x < W; ++x)
            {
                if(!(x & 3))
                {
                    const double t = clock();
                    if((improved && t > timeLimitSuccess) || t > timeLimitOtherwise) goto abortAlgorithm;
                }

                improved |= optimizeLocalBlock(x, x + w - 1, spreadY);
            }
        }

    abortAlgorithm:
		return *superFast.p_score > startScore;
	}

	bool adjustSingleBlocks(int maxBlockWidth,
							int maxOffset,
							const Clock* p_clock = 0,
							double timeLimit = 0.0) const
	{
		const ScoreType startScore = *superFast.p_score;

		for(int w = maxBlockWidth; w >= 1; --w)
		{
		    if(p_clock && (*p_clock)() > timeLimit) break;

			for(int x = 0; x < W; ++x)
			{
				bool yMinChecked = false;
				bool yMaxChecked = false;

				int minBorder = superFast.p_border[x];
				int maxBorder = minBorder;
				for(int i = x + 1; i < x + w; ++i)
				{
					const int y = superFast.p_border[i & (W - 1)];
					minBorder = std::min(minBorder, y);
					maxBorder = std::max(maxBorder, y);
				}

				int offsetDone = 0;
				for(int o = 0; o < 2 * maxOffset; ++o)
				{
					int offset = o >> 1;
					if(o & 1) offset = -offset;
					const int relativeOffset = offset - offsetDone;

                    if(yMinChecked && maxBorder + relativeOffset <= 1) offsetDone = offset;
                    else if(yMaxChecked && minBorder + relativeOffset >= (W / 2) - 1) offsetDone = offset;
                    else
                    {
                        bool improved;
                        if(maxBorder + relativeOffset <= (W / 2) - 1 && minBorder + relativeOffset >= 1) improved = superFastChangeBlockRelativeNoClip(x, x + w - 1, relativeOffset);
                        else improved = superFastChangeBlock(x, x + w - 1, relativeOffset, true);

                        if(improved)
                        {
                            offsetDone = offset;
                            if(maxBorder + relativeOffset <= 1) yMinChecked = true;
                            else if(minBorder + relativeOffset >= (W / 2) - 1) yMaxChecked = true;
                            minBorder = std::max(1, minBorder + relativeOffset);
                            maxBorder = std::min((W / 2) - 1, maxBorder + relativeOffset);
                        }
					}
				}
			}
		}

		return *superFast.p_score > startScore;
	}

    MapIntType map[WIDTH][HEIGHT];
    int scaleFactor;
    int scaleFactorSq;
    int scaleFactorLog;
    int scaleFactorSqLog;
    int landIntegral[WIDTH + 1][HEIGHT + 1];
    int16_t landColumnIntegral[HEIGHT + 1][WIDTH];
    int totalWater;
    int totalLand;
    long double twoOverTotalWaterTimesTotalLand;
};

template<int W> void scaleUp(const typename Helper<W>::BorderType& borderIn,
                             typename Helper<W * 2>::BorderType& borderOut)
{
    for(int x = 0; x < W * 2; ++x) borderOut[x] = borderIn[x >> 1] << 1;
}

int params[] = {
    15,   // 0
    8,    // 1
    31,   // 2
    10,   // 3
    63,   // 4
    16,   // 5
    9,    // 6
    61,   // 7
    24,   // 8
    9,    // 9
    67,   // 10
    10,   // 11
    14,   // 12
    108,  // 13
    12,   // 14
    12,   // 15
    22,   // 16
    4,    // 17
    2,    // 18
    34,   // 19
    4,    // 20
    24,   // 21
    14,   // 22
    86,   // 23
    8,    // 24
    43    // 25
};

void dividePlanet(const PlanetMap& planetMap,
                  Border& outBorder,
                  const Clock& clock)
{
    Helper<512> h512(planetMap);
    Helper<256> h256(h512);
    Helper<128> h128(h256);
    Helper<64> h64(h128);
    Helper<32> h32(h64);
    Helper<16> h16(h32);
    Helper<8> h8(h16);

    Helper<8>::BorderType border8;
    ScoreType bestScore = 0.0;

    int bestBorder8Combo = -1;
    for(int i = 0; i < 6561; ++i)
    {
        border8[0] = 1 + i % 3;
        border8[1] = 1 + (i / 3) % 3;
        border8[2] = 1 + (i / 9) % 3;
        border8[3] = 1 + (i / 27) % 3;
        border8[4] = 1 + (i / 81) % 3;
        border8[5] = 1 + (i / 243) % 3;
        border8[6] = 1 + (i / 729) % 3;
        border8[7] = 1 + (i / 2187) % 3;
        const ScoreType score = h8.quickEval(border8);
        if(score SCORE_CMP_ bestScore)
        {
            bestScore = score;
            bestBorder8Combo = i;
        }
    }

    for(int i = 0; i < 8; ++i)
    {
        border8[i] = 1 + bestBorder8Combo % 3;
        bestBorder8Combo /= 3;
    }

    Helper<16>::BorderType border16;
    Helper<16>::BorderType startBorder16;
    scaleUp<8>(border8, border16);
    memcpy(startBorder16, border16, sizeof(Helper<16>::BorderType));

    int bestBorder16Bits = 0;
	for(int i = 0; i < (1 << 16); ++i)
    {
		border16[0] = startBorder16[0] - (i & 1);
		border16[1] = startBorder16[1] - ((i >> 1) & 1);
		border16[2] = startBorder16[2] - ((i >> 2) & 1);
		border16[3] = startBorder16[3] - ((i >> 3) & 1);
		border16[4] = startBorder16[4] - ((i >> 4) & 1);
		border16[5] = startBorder16[5] - ((i >> 5) & 1);
		border16[6] = startBorder16[6] - ((i >> 6) & 1);
		border16[7] = startBorder16[7] - ((i >> 7) & 1);
		border16[8] = startBorder16[8] - ((i >> 8) & 1);
		border16[9] = startBorder16[9] - ((i >> 9) & 1);
		border16[10] = startBorder16[10] - ((i >> 10) & 1);
		border16[11] = startBorder16[11] - ((i >> 11) & 1);
		border16[12] = startBorder16[12] - ((i >> 12) & 1);
		border16[13] = startBorder16[13] - ((i >> 13) & 1);
		border16[14] = startBorder16[14] - ((i >> 14) & 1);
		border16[15] = startBorder16[15] - ((i >> 15) & 1);

        const ScoreType score = h16.quickEval(border16);
        if(score SCORE_CMP_ bestScore)
        {
            bestScore = score;
            bestBorder16Bits = i;
        }
    }

    for(int x = 0; x < 16; ++x)
        border16[x] = startBorder16[x] - ((bestBorder16Bits >> x) & 1);

    h16.initSuperFastEval(border16, &bestScore);
    h16.optimize();
	if(h16.adjustSingleBlocks(params[0], params[1])) h16.optimize();

    Helper<32>::BorderType border32;
    scaleUp<16>(border16, border32);
	h32.initSuperFastEval(border32, &bestScore);
	h32.optimize();
	if(h32.adjustSingleBlocks(params[2], params[3])) h32.optimize();

    Helper<64>::BorderType border64;
    scaleUp<32>(border32, border64);
	h64.initSuperFastEval(border64, &bestScore);
	h64.optimize();
	if(h64.adjustSingleBlocks(params[4], params[5])) h64.optimize();
	h64.optimizeLocalBinary2(1 << params[6]);

    Helper<128>::BorderType border128;
    scaleUp<64>(border64, border128);
	h128.initSuperFastEval(border128, &bestScore);
	h128.optimize();
	if(h128.adjustSingleBlocks(params[7], params[8])) h128.optimize();
	h128.optimizeLocalBinary2(1 << params[9]);

    Helper<256>::BorderType border256;
    scaleUp<128>(border128, border256);
	h256.initSuperFastEval(border256, &bestScore);
	h256.optimize();
	if(h256.adjustSingleBlocks(params[10], params[11])) h256.optimize();
	h256.optimizeLocalBinary2(1 << params[12]);

    scaleUp<256>(border256, outBorder);
	h512.initSuperFastEval(outBorder, &bestScore);
	h512.optimize();
	if(h512.adjustSingleBlocks(params[13], params[14])) h512.optimize();

	const double firstTimeLimit = 0.997;
	const double finalTimeLimit = 0.999;

	while(clock() < firstTimeLimit)
	{
		const ScoreType startScore = bestScore;

		h512.optimizeLocalBinary2(1 << params[15], &clock, firstTimeLimit);
		h512.adjustSingleBlocks(params[16], params[17], &clock, firstTimeLimit);
		h512.optimizeLocalBlocks2(params[18], params[19], params[20], params[21], &clock, firstTimeLimit);

		if(bestScore == startScore) break;

		h512.optimize(INT_MAX, &clock, firstTimeLimit);
	}

	if(h512.optimizeLocalBlocks3(params[22], params[23], params[24], params[25], clock, finalTimeLimit - 0.002, finalTimeLimit))
        h512.optimize(INT_MAX, &clock, finalTimeLimit);

	const int halfStep = params[24] / 2;
	h512.optimizeLocalBlocks2(params[22] + halfStep, params[23] + halfStep, params[24], params[25], &clock, finalTimeLimit);
}

#undef SCORE_CMP_
#undef COMPUTE_SCORE_

}
