// spieleprogrammierer.de
// Programmier-Contest #14: "Planetenteilung"
//
// Alle Informationen unter:
// http://www.spieleprogrammierer.de/index.php?page=Thread&threadID=16101

#define WIN32

#if defined(WIN32)
#include <windows.h>
#undef min
#undef max
#endif

#include <cassert>
#include <cmath>
#include <cstdio>
#include <ctime>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <vector>
#include "LodePNG.h"

// ********************************************************************************

const int MAP_WIDTH = 512;  // Breite der Planetenkarte
const int MAP_HEIGHT = 256; // Hhe der Planetenkarte

// Array-Typ fr eine Planetenkarte
typedef bool PlanetMap[MAP_WIDTH][MAP_HEIGHT];

// Array-Typ fr eine Grenzlinie
typedef int Border[MAP_WIDTH];

inline int absInt(int x)
{
    return x < 0 ? -x : x;
}

// Mini-Klasse, um die abgelaufene Zeit abzufragen
class Clock
{
public:
#if defined(WIN32)
    Clock()
    {
        LARGE_INTEGER f;
        QueryPerformanceFrequency(&f);
        conv = 1.0 / f.QuadPart;
        reset();
    }

    void reset()
    {
        QueryPerformanceCounter(&startTime);
    }

    double operator () () const
    {
        LARGE_INTEGER t;
        QueryPerformanceCounter(&t);
        return (t.QuadPart - startTime.QuadPart) * conv;
    }
#else
    Clock() : startTime(clock()), conv(1.0 / CLOCKS_PER_SEC) { }
    void reset() { startTime = clock(); }
    double operator () () const { return (clock() - startTime) * conv; }
#endif
private:
#if defined(WIN32)
    LARGE_INTEGER startTime;
    long double conv;
#else
    time_t startTime;
    double conv;
#endif
};

// ********************************************************************************
// Bewertungsfunktion, die einem Grenzverlauf eine Punktzahl zuweist.
// Siehe Kommentare in der Implementierung fr eine Erklrung der Berechnung.
// ********************************************************************************

struct Result
{
    Result()
        : score(0.0)
        , borderScore(0.0)
        , divisionScore(0.0)
    {
    }

    double score;
    double borderScore;
    double divisionScore;
};

Result computeScore(const PlanetMap& planetMap,
                    const Border& border)
{
    // zunchst die Lsung auf Gltigkeit prfen
    for(int x = 0; x < MAP_WIDTH; ++x)
	{
        if(border[x] <= 0 || border[x] >= MAP_HEIGHT)
		{
			char msg[256];
			sprintf(msg, "Ungltiger Grenzverlauf! border[%d] = %d", x, border[x]);
            throw std::runtime_error(msg);
		}
	}

    // Die Bewertung ergibt sich aus zwei Teilen:
    // 1.) "Lnge" der Grenze (s.u.)
    // 2.) Qualitt der Aufteilung (s.u.)

    // Zuerst wird die "Lnge" der Grenze ermittelt.
    // Jede y-nderung dy der Grenze von x nach x+1 (und von x=511 nach x=0)
    // fhrt zu einem Punktabzug von dy/2048.
    // Negative Punktzahlen sind mglich!
    // Eine in dieser Hinsicht optimale Grenze ist einfach eine horizontale Linie.
    // Eine solche Grenze bringt genau 1 Punkt.

    int borderLength = 0;
    for(int x = 0; x < MAP_WIDTH - 1; ++x) borderLength += absInt(border[x] - border[x + 1]);
    borderLength += absInt(border[MAP_WIDTH - 1] - border[0]);
    const double borderScore = 1.0 - borderLength / 2048.0;

    // Nun wird die Qualitt der Aufteilung des Planeten bestimmt.
    // Diese besteht aus zwei Faktoren:
    //
    // 1.) Wasser(Wasserbewohner) / Wasser(Gesamt)
    //     = Anteil des nutzbaren Wassers am Gesamtvorkommen auf dem Planeten
    //       (Wasser, das den Landbewohnern zugeteilt wird, gilt als verschwendet)
    //
    // 2.) Land(Landbewohner) / Land(Gesamt)
    //     = Anteil des nutzbaren Lands am Gesamtvorkommen auf dem Planeten
    //       (Land, das den Wasserbewohnern zugeteilt wird, gilt als verschwendet)
    //
    // Diese Faktoren werden miteinander multipliziert
    // und ergeben die zweite Teilpunktzahl.
    // Das theoretische Maximum ist 1, nmlich wenn die eine Hlfte des Planeten
    // zu 100% aus Wasser besteht und die andere zu 100% aus Land.

    // bestimmen, wie viel Wasser/Land nrdlich und sdlich der Grenze vorkommen
    int waterNorth = 0;
    int waterSouth = 0;
    int landNorth = 0;
    int landSouth = 0;
    for(int x = 0; x < MAP_WIDTH; ++x)
    {
        // nrdliche Seite der Grenze
        int landNorthInColumn = 0;
        for(int y = 0; y < border[x]; ++y) landNorthInColumn += planetMap[x][y] ? 1 : 0;
        landNorth += landNorthInColumn;
        waterNorth += border[x] - landNorthInColumn;

        // sdliche Seite der Grenze
        int landSouthInColumn = 0;
        for(int y = border[x]; y < MAP_HEIGHT; ++y) landSouthInColumn += planetMap[x][y] ? 1 : 0;
        landSouth += landSouthInColumn;
        waterSouth += (MAP_HEIGHT - border[x]) - landSouthInColumn;
    }

    // Gesamtvorkommen berechnen
    const int waterTotal = waterNorth + waterSouth;
    const int landTotal = landNorth + landSouth;

    // Wir wissen noch nicht, welche Bewohner auf welcher Seite leben sollen.
    // Darum probieren wir beide Mglichkeiten und whlen dann die bessere.
    const double divisionScore1 = (waterNorth / static_cast<double>(waterTotal)) * (landSouth / static_cast<double>(landTotal));
    const double divisionScore2 = (waterSouth / static_cast<double>(waterTotal)) * (landNorth / static_cast<double>(landTotal));
    const double bestDivisionScore = std::max(divisionScore1, divisionScore2);

    // Die endgltige Punktzahl ergibt sich:
    // - zu 1/3 aus der Punktzahl fr die Lnge der Grenze
    // - zu 2/3 aus der Punktzahl fr die Qualitt der Aufteilung

    Result r;
    r.score = (borderScore + 2.0 * bestDivisionScore) / 3.0;
    r.borderScore = borderScore;
    r.divisionScore = bestDivisionScore;
    return r;
}

// ********************************************************************************
// Ldt eine Planetenkarte aus einer PNG-Datei (512x256 Pixel).
// Blaue Pixel stellen Wasser dar, gelbe Pixel stellen Land dar.
// ********************************************************************************

void loadPlanetMapFromPNG(const std::string& filename,
                          std::vector<unsigned char>& outImage,
                          PlanetMap& outPlanetMap)
{
    // Bild laden
    std::vector<unsigned char> buffer;
    LodePNG::loadFile(buffer, filename);
    LodePNG::Decoder decoder;
    decoder.decode(outImage, buffer);

    // in eine Planetenkarte umwandeln
    std::vector<unsigned char>::const_iterator imageIt = outImage.begin();
    for(int y = 0; y < MAP_HEIGHT; ++y)
    {
        for(int x = 0; x < MAP_WIDTH; ++x)
        {
            const int r = *(imageIt++);
            const int g = *(imageIt++);
            const int b = *(imageIt++);
            ++imageIt;
            if(r == 0 && g == 0 && b == 255) outPlanetMap[x][y] = false;
            else if(r == 255 && g == 255 && b == 0) outPlanetMap[x][y] = true;
            else throw std::runtime_error("Nur blaue und gelbe Pixel sind erlaubt!");
        }
    }
}

// ********************************************************************************
// Einbindung der Algorithmen
// ********************************************************************************

#include "Algorithmen/Aremion.h"
#include "Algorithmen/Darkrel.h"
#include "Algorithmen/David_Scherfgen.h"
#include "Algorithmen/Helmut.h"
#include "Algorithmen/Wuempftlbruempftl.h"

typedef void (* DividePlanetFuncPtr)(const PlanetMap&, Border&, const Clock&);

const int NUM_PLANETS = 5;

Result bestScoreForPlanet[NUM_PLANETS];
Result bestAvgScoreForPlanet[NUM_PLANETS];

struct Submission
{
	char author[64];
	DividePlanetFuncPtr p_function;
	int numBestSolutions;
	Result avgScore;
	double avgTime;
	Result avgScoreForPlanet[NUM_PLANETS];
	double avgTimeForPlanet[NUM_PLANETS];
	int numErrorsForPlanet[NUM_PLANETS];
	int numErrors;
};

Submission submissions[] =
{
	{ "Aremion",			&Aremion::dividePlanet },
	{ "Darkrel",            &Darkrel::dividePlanet },
	{ "David Scherfgen",	&David_Scherfgen::dividePlanet },
	{ "Helmut",				&Helmut::dividePlanet },
	{ "Wuempftlbruempftl",  &Wuempftlbruempftl::dividePlanet }
};

const int NUM_SUBMISSIONS = static_cast<int>(sizeof(submissions) / sizeof(submissions[0]));

// ********************************************************************************
// Auswertungsprogramm
// ********************************************************************************

int main()
{
#if defined(WIN32)
	// hchste Prioritt fr den Prozess
	SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
#endif

    std::cout << "Return druecken, um zu starten!" << std::endl;
    std::cin.get();

    // Anzahl der Durchlufe pro Planet, ber die gemittelt wird
    const int NUM_PASSES = 5;

    for(int i = 0; i < NUM_SUBMISSIONS; ++i)
        submissions[i].numBestSolutions = 0,
        submissions[i].avgScore = Result(),
        submissions[i].avgTime = 0.0,
        submissions[i].numErrors = 0;

    for(int i = 0; i < NUM_PLANETS; ++i)
    {
        std::cout << "Teste Planet " << i << " ..." << std::endl;
        bestScoreForPlanet[i] = Result();
        bestAvgScoreForPlanet[i] = Result();

        char planetName[64];
        sprintf(planetName, "Planet%03d", i);

		// Karte laden
		char filename[64];
		sprintf(filename, "Planeten/%s.png", planetName);
		std::vector<unsigned char> planetImage;
		PlanetMap planetMap;
		loadPlanetMapFromPNG(filename, planetImage, planetMap);

        // HTML-Ausgabedatei ffnen
        char htmlFilename[64];
        sprintf(htmlFilename, "Ausgabe/%s.html", planetName);
        std::ofstream outHTML(htmlFilename);
        outHTML << "<html><head><title>Auswertung des sppro.de-Contests #14</title><link rel=\"stylesheet\" type=\"text/css\" href=\"../Auswertung.css\" /><script src=\"../Auswertung.js\" type=\"text/javascript\"></script><script src=\"../SortTable.js\" type=\"text/javascript\"></script></head><body>";
        outHTML << "<h1>Auswertung des sppro.de-Contests #14</h1>";
        outHTML << "<p>" << NUM_PLANETS << " Planeten, " << NUM_PASSES << " Durchg&auml;nge pro Planet.<br />Tabellen sind <strong>sortierbar</strong>!</p>";
        outHTML << "<h2>" << planetName << "</h2>";
        outHTML << "<p id=\"status\">Originalkarte</p>";
        outHTML << "<p><img id=\"image\" src=\"../Planeten/" << planetName << ".png\" onclick=\"return showOriginal('" << planetName << "');\" /></p>";
        outHTML << "<p>Fahre mit der Maus &uuml;ber einen Namen, um dessen L&ouml;sung anzuzeigen.<br />Klicke auf das Bild, um zur&uuml;ckzusetzen.</p>";
        outHTML << "<table class=\"sortable\"><thead><tr><th>Name</th><th>Punktzahl</th><th>Grenzverlauf</th><th>Aufteilung</th><th>Zeit (ms)</th><th>Fehler/Timeouts</th></tr></thead><tbody>";

		for(int j = 0; j < NUM_SUBMISSIONS; ++j)
		{
		    std::cout << std::setw(20) << submissions[j].author << " ";

            Border border[NUM_PASSES];
            Clock clock;

			// Jeder bekommt einen Aufruf als "Warm-up", der nicht zhlt!
			// Beim allerersten Aufruf machen wir sogar 10 Warm-ups.
			const int numWarmUps = (i == 0 ? 10 : 1);
			for(int k = 0; k < numWarmUps; ++k)
			{
			    clock.reset();
			    //(*submissions[j].p_function)(planetMap, border[0], clock);
			    std::cout << "warmup3210"[k];
			}

			Result scoreForPlanetPass[NUM_PASSES];
			double timeForPlanetPass[NUM_PASSES];
			Result avgScoreForPlanet;
			double avgTimeForPlanet = 0.0;
			submissions[j].numErrorsForPlanet[i] = 0;

			for(int k = 0; k < NUM_PASSES; ++k)
			{
				try
				{
					// Aber jetzt zhlt's!
					// Grenze und Uhr zurcksetzen, dann Algorithmus aufhren und Zeit stoppen.
					scoreForPlanetPass[k] = Result();
					for(int x = 0; x < MAP_WIDTH; ++x) border[k][x] = 0;
					clock.reset();
					(*submissions[j].p_function)(planetMap, border[k], clock);
					const double timeNeeded = clock();
					timeForPlanetPass[k] = timeNeeded;
					avgTimeForPlanet += timeNeeded;

					// Lsung bewerten
					const Result score = computeScore(planetMap, border[k]);
					if(timeNeeded <= 1.001)
					{
					    scoreForPlanetPass[k] = score;
					    avgScoreForPlanet.score += score.score;
					    avgScoreForPlanet.borderScore += score.borderScore;
					    avgScoreForPlanet.divisionScore += score.divisionScore;
					    if(score.score > bestScoreForPlanet[i].score) bestScoreForPlanet[i] = score;
					}
					else ++submissions[j].numErrorsForPlanet[i];
				}
				catch(std::exception& ex)
				{
					std::cerr << "Exception wurde geworfen: " << ex.what() << std::endl;
					++submissions[j].numErrorsForPlanet[i];
				}

				std::cout << '.';
			}

            avgScoreForPlanet.score /= NUM_PASSES;
            avgScoreForPlanet.borderScore /= NUM_PASSES;
            avgScoreForPlanet.divisionScore /= NUM_PASSES;
            avgTimeForPlanet /= NUM_PASSES;
            if(avgScoreForPlanet.score > bestAvgScoreForPlanet[i].score) bestAvgScoreForPlanet[i] = avgScoreForPlanet;
            submissions[j].avgScore.score += avgScoreForPlanet.score;
            submissions[j].avgScore.borderScore += avgScoreForPlanet.borderScore;
            submissions[j].avgScore.divisionScore += avgScoreForPlanet.divisionScore;
            submissions[j].avgTime += avgTimeForPlanet;
			submissions[j].avgScoreForPlanet[i] = avgScoreForPlanet;
			submissions[j].avgTimeForPlanet[i] = avgTimeForPlanet;
			submissions[j].numErrors += submissions[j].numErrorsForPlanet[i];

			// Grenzverlauf suchen, der dem Durchschnitt der Lsungen dieses Kandidaten am nchsten kommt
			double bestDiff = 1.0;
			int bestPass = -1;
			for(int k = 0; k < NUM_PASSES; ++k)
			{
			    if(scoreForPlanetPass[k].score != 0.0)
			    {
			        const double diff = std::abs(scoreForPlanetPass[k].score - avgScoreForPlanet.score);
                    if(diff < bestDiff) bestDiff = diff, bestPass = k;
			    }
			}

            if(bestPass != -1)
            {
                // Bild generieren
                std::vector<unsigned char> borderImage(planetImage);
                for(int x = 0; x < MAP_WIDTH; ++x)
                {
                    for(int y = border[bestPass][x]; y < MAP_HEIGHT; ++y)
                    {
                        const int base = y * MAP_WIDTH * 4 + x * 4;
                        if(planetMap[x][y]) borderImage[base] = 255, borderImage[base + 1] = 127, borderImage[base + 2] = 0;
                        else borderImage[base] = 127, borderImage[base + 1] = 0, borderImage[base + 2] = 255;
                    }
                }
                char borderImageFilename[128];
                sprintf(borderImageFilename, "Ausgabe/%s - %s.png", planetName, submissions[j].author);
                LodePNG::Encoder encoder;
                std::vector<unsigned char> buffer;
                encoder.encode(buffer, &borderImage[0], MAP_WIDTH, MAP_HEIGHT);
                LodePNG::saveFile(buffer, borderImageFilename);

                outHTML << "<tr><td><a href=\"#\" onclick=\"return false;\" onmouseover=\"return showSolution('" << planetName << "', '" << submissions[j].author << "');\">" << submissions[j].author << "</a></td><td>"
                        << scoreForPlanetPass[bestPass].score << "</td><td>" << scoreForPlanetPass[bestPass].borderScore << "</td><td>" << scoreForPlanetPass[bestPass].divisionScore << "</td><td>"
                        << static_cast<int>(floor(timeForPlanetPass[bestPass] * 1000.0 + 0.5)) << "</td><td>" << submissions[j].numErrorsForPlanet[i] << "</td></tr>";
            }

			std::cout << " S=" << avgScoreForPlanet.score << ", S_avg=" << (submissions[j].avgScore.score / (i + 1)) << ", T=" << avgTimeForPlanet;
			if(submissions[j].numErrorsForPlanet[i]) std::cout << ", " << submissions[j].numErrorsForPlanet[i] << " Fehler/Timeouts!";
			std::cout << std::endl;
		}

		outHTML << "</tbody></table>";

		// Wer war hier am besten? (Knnen auch mehrere sein!)
		for(int j = 0; j < NUM_SUBMISSIONS; ++j)
            if(submissions[j].avgScoreForPlanet[i].score == bestAvgScoreForPlanet[i].score)
                ++submissions[j].numBestSolutions;

		std::cout << std::endl;
    }

    std::cout << std::endl;
    std::cout << "Statistiken:" << std::endl;
    std::cout << "============" << std::endl << std::endl;

    Result theoreticalBestAvgScore;
    for(int i = 0; i < NUM_PLANETS; ++i)
        theoreticalBestAvgScore.score += bestScoreForPlanet[i].score,
        theoreticalBestAvgScore.borderScore += bestScoreForPlanet[i].borderScore,
        theoreticalBestAvgScore.divisionScore += bestScoreForPlanet[i].divisionScore;
    theoreticalBestAvgScore.score /= NUM_PLANETS;
    theoreticalBestAvgScore.borderScore /= NUM_PLANETS;
    theoreticalBestAvgScore.divisionScore /= NUM_PLANETS;
    std::cout << "Theoretisch erreichbare durchschnittliche Punktzahl:" << std::endl << theoreticalBestAvgScore.score << std::endl << std::endl;

    for(int i = 0; i < NUM_SUBMISSIONS; ++i)
    {
        submissions[i].avgScore.score /= NUM_PLANETS;
        submissions[i].avgScore.borderScore /= NUM_PLANETS;
        submissions[i].avgScore.divisionScore /= NUM_PLANETS;
        submissions[i].avgTime /= NUM_PLANETS;

        std::cout << submissions[i].author << ":" << std::endl;
        std::cout << "Durchschnittliche Punktzahl: " << submissions[i].avgScore.score << std::endl;
        std::cout << "Durchschnittliche Zeit:      " << submissions[i].avgTime << std::endl;
        std::cout << "Anzahl bester Loesungen:     " << submissions[i].numBestSolutions << std::endl;
        std::cout << std::endl;
    }

    // ********************************************************************************

    // HTML-Ausgabe
    std::ofstream outHTML("Auswertung.html");
    outHTML << "<html><head><title>Auswertung des sppro.de-Contests #14</title><link rel=\"stylesheet\" type=\"text/css\" href=\"Auswertung.css\" /><script src=\"SortTable.js\" type=\"text/javascript\"></script></head><body>";
    outHTML << "<h1>Auswertung des sppro.de-Contests #14</h1>";
    outHTML << "<p>" << NUM_PLANETS << " Planeten, " << NUM_PASSES << " Durchg&auml;nge pro Planet.<br />Tabellen sind <strong>sortierbar</strong>!</p>";
    outHTML << "<h2>&Uuml;bersicht</h2>";
    outHTML << "<table class=\"sortable\"><thead><tr><th>Name</th><th>&#216; Punktzahl</th><th>&#216; Grenzverlauf</th><th>&#216; Aufteilung</th><th>&#216; Zeit (ms)</th><th>Anzahl bester L&ouml;sungen</th><th>Fehler/Timeouts</th></tr></thead><tbody>";
    for(int i = 0; i < NUM_SUBMISSIONS; ++i) outHTML << "<tr><td>" << submissions[i].author << "</td><td>" << submissions[i].avgScore.score << "</td><td>" << submissions[i].avgScore.borderScore << "</td><td>" << submissions[i].avgScore.divisionScore << "</td><td>" << static_cast<int>(floor(submissions[i].avgTime * 1000.0 + 0.5)) << "</td><td>" << submissions[i].numBestSolutions << "</td><td>" << submissions[i].numErrors << "</td></tr>";
    outHTML << "<tr style=\"font-style: italic;\"><td>&quot;Optimaler&quot; Algorithmus</td><td>" << theoreticalBestAvgScore.score << "</td><td>" << theoreticalBestAvgScore.borderScore << "</td><td>" << theoreticalBestAvgScore.divisionScore << "</td><td>?</td><td>" << NUM_PLANETS << "</td><td>0</td></tr></tbody></table>";
    outHTML << "<h2>Punktzahlen bei den einzelnen Planeten</h2>";
    outHTML << "<p>Klicke auf eine Planetennummer, um die L&ouml;sungen anzuzeigen.</p>";
    outHTML << "<table class=\"sortable\"><thead><tr><th>Planet</th>";
    for(int i = 0; i < NUM_SUBMISSIONS; ++i) outHTML << "<th>" << submissions[i].author << "</th>";
    outHTML << "<th style=\"font-style: italic;\">&quot;Optimaler&quot; Algorithmus</th></tr></thead><tbody>";
    for(int i = 0; i < NUM_PLANETS; ++i)
    {
        char planetName[64];
        sprintf(planetName, "%03d", i);
        outHTML << "<tr><td><a href=\"Ausgabe/Planet" << planetName << ".html\">" << planetName << "</a></td>";
        for(int j = 0; j < NUM_SUBMISSIONS; ++j)
        {
            const char* p_td = (submissions[j].avgScoreForPlanet[i].score == bestScoreForPlanet[i].score) ? "<td class=\"bestscore\">" : "<td>";
            outHTML << p_td << submissions[j].avgScoreForPlanet[i].score << "</td>";
        }
        outHTML << "<td style=\"font-style: italic;\">" << bestScoreForPlanet[i].score << "</td></tr>";
    }
    outHTML << "</tbody></table></body></html>";

    return 0;
}
